#ifndef __WORDBLOX_H
#define __WORDBLOX_H

#include <gd.h>
#include <gdfontmb.h>
#include <gdfontg.h>
#include "constantstrings.h"

#define MAX_PUZZLE_SIZE 20
#define MAX_CLUE_LENGTH 150

/* Enum to define terminal colours */
enum COLOR {
	BLACK = 0,
	RED= 1,
	GREEN=2,
	YELLOW=3,
	BLUE=4,
	MAGENTA=5,
	CYAN=6,
	WHITE=7
};

enum ATTR {
	NORMAL = 23,
	BOLD=1
};

enum ORIENTATION {
	ACROSS=1,
	DOWN=2
};

typedef char String[MAX_CLUE_LENGTH];

/* The main puzzle struct type */
typedef struct {
	char chars[MAX_PUZZLE_SIZE][MAX_PUZZLE_SIZE];
	int start_across_word[MAX_PUZZLE_SIZE][MAX_PUZZLE_SIZE];
	int start_down_word[MAX_PUZZLE_SIZE][MAX_PUZZLE_SIZE];
	String clue_across[MAX_PUZZLE_SIZE][MAX_PUZZLE_SIZE];
	String clue_down[MAX_PUZZLE_SIZE][MAX_PUZZLE_SIZE];
	int grid_size;
	bool grid_frozen;
} Puzzle;

/* get a number from the user */
int get_num ()
{
	char s[5];
	fgets (s, 5, stdin);
	int n = atoi (s);
	return n;
}

/* Output the grid to image - if answerkey is true export filled grid */
void export_grid_image  (Puzzle *p, const char *filename, bool answerkey) 
{
	int img_size = p->grid_size * 40;
	FILE * outfile = fopen (filename, "wb");
	if (outfile == NULL)
	{
		fprintf (stderr, "%s\n", ERROR_WRITING_FILE);
		exit (1);
	}
	
	gdImagePtr img = gdImageCreate (img_size, img_size);
	int white = gdImageColorAllocate (img, 255,255,255);
	int black = gdImageColorAllocate (img, 0, 0, 0);
	int blue = gdImageColorAllocate (img, 0, 0, 216);
	gdFontPtr sm_fnt = gdFontGetMediumBold ();
	gdFontPtr lg_fnt = gdFontGetGiant ();
	
	for (int i = 0; i < p->grid_size; i ++)
	{
		for (int j = 0; j < p->grid_size; j++)
		{
			/* if it is a block, draw the black square */
			if (p->chars[i][j] == '#')
				gdImageFilledRectangle (img, j*40, i*40, j*40+40, 
											i*40+40,black);
			else
			{
				/* draw a regular square */
				gdImageRectangle (img, j*40, i*40, j*40+40, 
									i*40+40, black);
				
				/* print the numers, if it is either start across word or 
				a down word */
				if (p->start_across_word[i][j] != -1 || 
						p->start_down_word[i][j] != -1)
				{
					if (p->start_across_word[i][j] != -1)
					{
						char str[5];
						sprintf (str, "%d", p->start_across_word[i][j]);
						gdImageString (img,  sm_fnt, j*40+2, i*40+2,
									(unsigned char *)str, blue);   
					}
					else 
					{
						char str[5];
						sprintf (str, "%d", p->start_down_word[i][j]);
						gdImageString (img,  sm_fnt, j*40+2, i*40+2,
									(unsigned char *)str, blue);					
					}
				}
				/* if answerkey is true, draw the character in the cell */
				if (answerkey)
				{
					gdImageChar (img, lg_fnt, j*40+15, i*40+15, 
						p->chars[i][j], black);
				}
			}
		}
	}
	
	gdImagePng (img, outfile);
	gdImageDestroy (img);
	fclose (outfile);
}

/* Set the terminal colour */
void set_color (enum COLOR fg, enum COLOR bg, enum ATTR at) {
	printf ("\x1B[%d;%d;%dm", fg+30, bg+40, at);
}

/* Reset the terminal colour */
void reset_color () {
	printf ("\x1B[0m");
}

/* check if previous row is blank or not */
bool prev_row_blank (Puzzle *p, int r, int c) 
{
	if (r == 0) return true;
	if (p->chars[r-1][c] == ' ' || p->chars[r-1][c] == '#') return true;
	return false;
}
/* check if next row is blank or not */
bool next_row_blank (Puzzle *p, int r, int c)
{
	if (r == p->grid_size - 1) return true;
	if (p->chars[r+1][c] == ' ' || p->chars[r+1][c] == '#') return true;
	return false;
}
/* check if previous col is blank or not */
bool prev_col_blank (Puzzle *p, int r, int c)
{
	if (c == 0) return true;
	if (p->chars[r][c-1] == ' ' || p->chars[r][c-1] == '#') return true;
	return false; 
}
/* check if the next col is blank or not */
bool next_col_blank (Puzzle *p, int r, int c)
{
	if (c == p->grid_size -1) return true;
	if (p->chars[r][c+1] == ' ' || p->chars[r][c+1] == '#') return true;
	return false;
}

/* unfreeze the grid - mak editing possible to change words */
void unfreeze_puzzle (Puzzle *p)
{
	for (int i = 0; i < p->grid_size; i ++)
	{
		for (int j = 0; j < p->grid_size; j ++)
		{
			if (p->chars[i][j] == '#')
				p->chars[i][j] = ' ';
			
			p->start_across_word[i][j] = -1;
			p->start_down_word[i][j] = -1;
		}
	}
	p->grid_frozen = false;
}

/* freeze the grid - make editing impossible because it finalizes the
   across and down words in the grid */
void freeze_puzzle (Puzzle *p)
{
	int word_num = 1;
	bool across_word_start, down_word_start;
	for (int i = 0; i < p->grid_size; i ++)
	{
		for (int j = 0; j < p->grid_size; j++)
		{
			across_word_start = false; 
			down_word_start = false;
			/* if it is a blank cell - cover it with a block */
			if (p->chars[i][j] == ' ')
				p->chars[i][j] = '#';
			/* it is not a blank cell - check all possibilities */
			else 
			{
				bool prev_row = prev_row_blank (p, i, j);
				bool next_row = next_row_blank (p, i, j);
				bool prev_col = prev_col_blank (p, i, j);
				bool next_col = next_col_blank (p, i, j);
				if (prev_row && ! next_row)
					down_word_start = true;
				if (prev_col && ! next_col)
					across_word_start = true;
			}
			
			if (across_word_start == true)
				p->start_across_word[i][j] = word_num;
			else 
				p->start_across_word[i][j] = -1;
			if (down_word_start == true)
				p->start_down_word[i][j] = word_num;
			else
				p->start_down_word[i][j] = -1;
			if (across_word_start == true || down_word_start == true)
				word_num ++;
		}
	}
	p->grid_frozen = true;
}

/* reset the entire grid */
void init_puzzle (Puzzle *p, int grid_size) 
{
	p->grid_size = grid_size;
	p->grid_frozen = false;
	for (int i = 0; i < p->grid_size; i ++)
	{
		for (int j = 0; j < p->grid_size; j ++) 
		{
			p->chars[i][j] = ' ';
			p->start_across_word[i][j] = -1;
			p->start_down_word[i][j] = -1;
			strcpy (p->clue_across[i][j], "");
			strcpy (p->clue_down[i][j], "");			
		}
	}
}

/* save the puzzle */
void save_puzzle (Puzzle *puzzle, const char* file) {
	FILE *outfile;
	outfile = fopen (file, "wb");
	if (outfile == NULL)
	{
		fprintf (stderr, "%s\n", ERROR_WRITING_FILE);
		exit (1);
	}
	fwrite (puzzle, sizeof (*puzzle), 1, outfile);
	fclose (outfile);
}

/* read the puzzle */
Puzzle load_puzzle (const char* file) {
	FILE *infile;
	Puzzle p;
	infile = fopen (file, "rb");
	if (infile == NULL)
	{
		fprintf (stderr, "%s\n", ERROR_READING_FILE);
		exit (1);
	}
	fread (&p, sizeof(p), 1, infile);
	fclose (infile);
	return p;
}

/* display the puzzle */ 
void print_puzzle (Puzzle *p) 
{
	printf ("\n");
	set_color (WHITE, CYAN, NORMAL);
	printf ("    ");
	for (int i = 0; i < p->grid_size; i ++)
		printf ("%3d", i);
	reset_color ();
	printf("\n");
	for (int i = 0; i < p->grid_size; i ++)
	{
		set_color (WHITE, CYAN, NORMAL);
		printf ("%3d ", i);
		for (int j = 0; j < p->grid_size; j ++) 
		{
			if (p->chars[i][j] == '#') {
				set_color (WHITE, BLACK, NORMAL);
				printf ("   ");
			}
			else 
			{
				if (p->start_across_word[i][j] != -1 || 
					p->start_down_word[i][j] != -1) 
				{
					set_color (BLUE, WHITE, NORMAL);
					if (p->start_across_word[i][j] != -1)
						printf ("%-2d", p->start_across_word[i][j]);
					else 
						printf ("%-2d", p->start_down_word[i][j]);
				}
				else
				{
					set_color (BLACK, WHITE,NORMAL);
					printf ("  ");
				}
					
				set_color (BLACK, WHITE, BOLD);
				printf ("%c", p->chars[i][j]);
			}
			reset_color ();
		}
		printf ("\n");
	}
	/* print the clues if set */
	if (p->grid_frozen == true) 
	{
		printf ("\x1B[1mACROSS - CLUES\x1B[0m\n");
		for (int i = 0; i < p->grid_size; i ++)
		{
			for (int j = 0; j < p->grid_size; j ++)
			{
				if (p->start_across_word[i][j] != -1)
				{
					printf ("%d - %s; ", p->start_across_word[i][j], 
								p->clue_across[i][j]);
				}
			}
		}
		printf ("\n\x1B[1mDOWN - CLUES\x1B[0m\n");
		for (int i = 0; i < p->grid_size; i ++)
		{
			for (int j = 0; j < p->grid_size; j ++)
			{
				if (p->start_down_word[i][j] != -1)
				{
					printf ("%d - %s; ", p->start_down_word[i][j], 
								p->clue_down[i][j]);
				}
			}
		}
		printf ("\n");	
	}
}

/* function to check if a word is valid or not */
bool is_valid_word (const char *word) 
{
	for (int i = 0; i < strlen (word); i ++)
		if (! isalpha (word[i])) 
			return false;
			
	return true;
}


/* function to set a clue for an across word */
bool set_clue (Puzzle *p, String clue, int index, enum ORIENTATION order) 
{
	for (int i = 0; i < p->grid_size; i ++)
	{
		for (int j = 0; j < p->grid_size; j ++)
		{
			if (order == ACROSS)
			{
				if (p->start_across_word[i][j] == index)
				{
					strcpy (p->clue_across[i][j], clue);
					return true;
				}
			}
			else if (order == DOWN)
			{
				if (p->start_down_word[i][j] == index)
				{
					strcpy (p->clue_down[i][j], clue);
					return true;
				}
			}		
		}
	}
	return false;
}

/* function to print a menu */
void print_menu (enum COLOR fg, enum COLOR bg, const char* title, 
					char **items, int num_items, int padding)
{
		/* clear screen */
		printf ("\e[1;1H\e[2J");
		set_color (fg, bg, NORMAL);
		printf ("\u2554");
		for (int i = 0; i < padding; i ++)
			printf ("\u2550");
		printf ("\u2557"); 
		reset_color (); printf ("\n");
		set_color (fg, bg, BOLD);
		printf ("\u2551%-*s\u2551", padding, title);
		reset_color (); printf ("\n");
		set_color (fg, bg, NORMAL);
		printf ("\u2560");
		for (int i = 0; i < padding; i ++)
			printf ("\u2550");
		printf ("\u2563");
		reset_color (); printf ("\n");
		for (int i = 0; i < num_items; i ++) 
		{
			set_color (fg, bg, NORMAL);
			printf ("\u2551%-*s\u2551", padding, items[i]);
			reset_color (); printf ("\n");
		}
		set_color (fg, bg, NORMAL);
		printf ("\u255A");
		for (int i = 0; i < padding; i ++)
			printf ("\u2550");
		printf ("\u255D");
		reset_color (); printf ("\n");
}

#endif
